import fuzzy
import fuzzy.storage.fcl.Reader

def process_operators(operator, verbose=False):
	if isinstance(operator, fuzzy.operator.Input.Input):
		if verbose: print "    Simple input", operator.adjective.getName(fuzzySystem), 
		MEMB = operator.adjective.membership
		if verbose: print "membership =", MEMB
		if MEMB==None:
			warnings.warn("input for "+operator.adjective.getName(self.fuzzySystem)[1]+" not specified")
		return MEMB
		
	elif isinstance(operator, fuzzy.operator.Compound.Compound):
		if verbose: print "    Multiple inputs: "			
		BEST = sys.float_info.max
		for inp in operator.inputs:
			MEMB = process_operators(inp, verbose=verbose)
			if MEMB==None: 
					warnings.warn("input for "+inp.adjective.getName(self.fuzzySystem)[1]+" not specified")
			elif MEMB<BEST:
				BEST = MEMB			
		return BEST
				


def get_singleton_value(adj):
	return adj.set.getCOG()

def return_all_output_variable_names(fuzzySystem):
	ret = {}
	for name, i in fuzzySystem.variables.items():
		if isinstance(i , fuzzy.OutputVariable.OutputVariable):
			ret[str(name)]=[0, 0]
	return ret

def actual_sugeno_inference(fuzzySystem, verbose=False):

	all_outputs = return_all_output_variable_names(fuzzySystem)		

	for name, rule in fuzzySystem.rules.items():
		if verbose: print " * Processing rule", name

		MEMB = 0
		DENO = 0
		NUME = 0

		SINGLETON = 0


		if isinstance(rule.adjective, fuzzy.Adjective.Adjective):
			if verbose: print "WARNING: Aggettivo output semplice, non supportato"
			pass
		elif isinstance(rule.adjective,list):
			
			if verbose: print "  * Processing list of OUTPUT adjectives..."

			# what output are we updating?
			for adj in rule.adjective:
				outlabel, outname = adj.getName(fuzzySystem)
				
				if verbose: print "   ", outlabel, outname, 
				SINGLETON = get_singleton_value(adj)
				if verbose: print "whose singleton is", SINGLETON

			# process operators (i.e., input nodes)
			MEMB = process_operators(rule.operator, verbose=verbose)
			if MEMB==None:
				# raise Exception("error: input for "+inp.adjective.getName(self.fuzzySystem)[1]+" not specified")
				# warnings.warn("input for "+inp.adjective.getName(self.fuzzySystem)[1]+" not specified")
				if verbose: print "WARNING: cannot calculate MAX membership value"
				pass
			else:
				DENO += MEMB	
				NUME += MEMB*SINGLETON
				if verbose: print "      Product =", MEMB*SINGLETON

		else:
			raise Exception("rule target not set.")

		all_outputs[outname][0]+=NUME
		all_outputs[outname][1]+=DENO

	if verbose: print all_outputs

	final = {}
	for out, valout in all_outputs.items():
		if valout[1]==0:
			final[out]=0
		else:
			final[out] = valout[0]/valout[1]
	return final

def sugeno_inference(my_input, fuzzySystem, verbose=False):
	"""
		Calculates the Sugeno rule of inference using the specified
		input values and updating the specified output values.
		Appearently, the Sugeno method is not implemented in pyfuzzy. 
		This method replaces pyfuzzy's calculate() method.
	"""

	fuzzySystem.reset()
	fuzzySystem.fuzzify(my_input)
	
	# the following replaces pyfuzzy's inference() method
	my_output = actual_sugeno_inference(fuzzySystem, verbose=verbose)

	# self.fuzzySystem.defuzzify(my_output)
	return my_output



if __name__ == '__main__':
	
	pass


import sys
import copy
import logging
import numpy as np
from fstpso import FuzzyPSO


class FFTPSO(FuzzyPSO):

    def __init__(self, alpha=0.01, logfile=None):
        super(FFTPSO, self).__init__(logfile=logfile)

        self.rollback_swarm = None
        self.g_particle_id = None

        if sys.version_info[0] < 3:
            self.maximum_integer = sys.maxint
        else:
            self.maximum_integer = sys.maxsize
        self.alpha = self.maximum_integer
        self.alpha_factor = alpha

    def _print_banner(self):
        print ("Fuzzy Time PSO 1.0")

    def Solve(self,
              funz,
              verbose=False,
              callback=None,
              dump_best_fitness=None,
              dump_best_solution=None,
              print_bar=True):

        self.update_alpha(verbose)
        self.update_rollback_swarm(verbose)

        logging.info('Launching optimization.')

        if verbose:
            print(" * Process started")

        while not self.TerminationCriterion(verbose=verbose):
            if funz is not None:
                funz(self)

            self.Iterate(verbose)
            self.rewind_swarm(verbose)

            if verbose:
                print(f"Completed iteration {self.Iterations}")
            else:
                if print_bar:
                    if self.Iterations in [self.MaxIterations // x for x in range(1, 10)]:
                        print(f" * %dth iteration out of %d completed. {self.Iterations, self.MaxIterations}", end="")
                        print("[%s%s]" %
                              (
                                  "#" * int(30 * self.Iterations / self.MaxIterations),
                                  " " * (30 - int(30 * self.Iterations / self.MaxIterations))
                              )
                              )

            # if a callback is specified, call it at regular intervals
            if callback is not None:
                interval = callback['interval']
                function = callback['function']
                if (self.Iterations - 1) % interval == 0:
                    function(self)
                    if verbose:
                        print(" * Callback invoked")

            # write the current best fitness
            if dump_best_fitness is not None:
                if self.Iterations == 1:
                    with open(dump_best_fitness, "w") as fo:
                        pass  # touch
                with open(dump_best_fitness, "a") as fo:
                    fo.write(str(self.G.CalculatedFitness) + "\n")

            # write the current best solution
            if dump_best_solution is not None:
                if self.Iterations == 1:
                    with open(dump_best_solution, "w") as fo:
                        pass  # touch
                with open(dump_best_solution, "a") as fo:
                    fo.write("\t".join(map(str, self.G.X)) + "\n")

        if verbose:
            print(f" * Optimization process terminated. Best solution found: {self.G.X}, "
                  f"with fitness {self.G.CalculatedFitness}")

        logging.info(f"Best solution: {str(self.G)}")
        logging.info(f"Fitness of best solution: {self.G.CalculatedFitness}")

        if self._discrete_cases is None:
            return self.G, self.G.CalculatedFitness
        else:
            return self.G, self.G.CalculatedFitness, self._best_discrete_sample

    def update_alpha(self, verbose):
        if type(self.alpha_factor) == float:
            if self.alpha_factor < 0.0 or self.alpha_factor > 1.0:
                self.alpha_factor = 0.05
            self.alpha = int(self.MaxIterations * self.alpha_factor)
        elif type(self.alpha_factor) == int:
            self.alpha = self.alpha_factor
        else:
            raise ValueError
        if verbose:
            print(f"* alpha is set to {self.alpha} iterations")

    def update_rollback_swarm(self, verbose):
        self.rollback_swarm = []
        for p in self.Solutions:
            self.rollback_swarm.append(copy.deepcopy(p))
        if verbose:
            print("Rollback swarm is updated")

    def rewind_swarm(self, verbose):
        if self.SinceLastGlobalUpdate >= self.alpha:
            print(" * Rewinding the swarm at the intial state")
            self.SinceLastGlobalUpdate = 0
            self.Solutions = []
            for p in self.rollback_swarm:
                self.Solutions.append(copy.deepcopy(p))

            self.Solutions[self.GIndex].X = self.NewGenerate([0] * len(self.Boundaries),
                                                             creation_method={'name': "uniform"})
            self.Solutions[self.GIndex].B = copy.deepcopy(self.Solutions[self.GIndex].X)
            self.Solutions[self.GIndex].V = list(np.zeros(len(self.Boundaries)))
            self.Solutions[self.GIndex].CalculatedFitness = self.maximum_integer

            self.determine_G()
            self.update_rollback_swarm(verbose=verbose)

    def determine_G(self):
        particles_fs = [(i, self.Solutions[i].CalculatedFitness) for i in range(len(self.Solutions))]
        particles_fs = sorted(particles_fs, key=lambda x: x[1])
        self.GIndex = particles_fs[0][0]
        self.G = copy.deepcopy(self.Solutions[self.GIndex])



from fstpso import FuzzyPSO	

def example_fitness( particle ):
	return sum(map(lambda x: x**2, particle))

if __name__ == '__main__':
	
	dims = 10
	FP = FuzzyPSO()
	FP.set_search_space( [[-10, 10]]*dims )	
	FP.set_fitness(example_fitness)	
	result =  FP.solve_with_fstpso(max_iter=100)
	print "Best solution:", result[0]
	print "Whose fitness is:", result[1]
